#include <stdio.h>
#include <tchar.h>
#include <iostream>
#include <math.h>

#include <driver_types.h>
#include <cuda_runtime_api.h>
#include <cutil.h>

#define ARRAY_SIZE 10000000
#define MA_PERIOD 35

extern "C" void SMAGold(float* pOut, float* pIn, float fPeriod, long lSize);
extern "C" void SMACuda(float* pOut, float* pIn, float fPeriod, long lSize);

float RandFloat(float low, float high);

int _tmain(int argc, _TCHAR* argv[])
{
	CUT_DEVICE_INIT();

	float* pSourceData = new float [ARRAY_SIZE];
	float* pDestDataGold = new float [ARRAY_SIZE];
	float* pDestDataCuda = new float [ARRAY_SIZE];

	srand(1234);
	for (int i = 0; i < ARRAY_SIZE; i++)
		pSourceData[i] = RandFloat(0, 100);

	double dCPUTime;
	double dGPUTime;
	unsigned int hTimer;
	CUT_SAFE_CALL(cutCreateTimer(&hTimer));

	_tprintf(_T("Running gold on CPU ...\n"));
	CUT_SAFE_CALL(cutStartTimer(hTimer));
	SMAGold(pDestDataGold, pSourceData, MA_PERIOD, ARRAY_SIZE);
	CUT_SAFE_CALL(cutStopTimer(hTimer));
	dCPUTime = cutGetTimerValue(hTimer);
    _tprintf(_T("CPU time: %f msec\n\n"), dCPUTime);

	_tprintf(_T("Running kernel on GPU ...\n"));
	CUT_SAFE_CALL(cutResetTimer(hTimer));
	CUT_SAFE_CALL(cutStartTimer(hTimer));
	SMACuda(pDestDataCuda, pSourceData, MA_PERIOD, ARRAY_SIZE);
	CUT_SAFE_CALL(cutStopTimer(hTimer));
	dGPUTime = cutGetTimerValue(hTimer);
    _tprintf(_T("GPU full time: %f msec\n\n"), dGPUTime);

	_tprintf(_T("Comparing results ...\n"));
	double dSumRef = 0.0;
	double dSumDelta = 0.0;
	double dMaxDelta = 0.0;
	for (int i = 0; i < ARRAY_SIZE; i++)
	{
		double dRef = pDestDataGold[i];
		double dDelta = fabs(pDestDataGold[i] - pDestDataCuda[i]);
		
		if (dDelta > dMaxDelta)
			dMaxDelta = dDelta;
		
		dSumDelta += dDelta;
		dSumRef   += dRef;
	}
	double dL1Norm = dSumDelta / dSumRef;
	_tprintf(_T("L1 norm: %E\n"), dL1Norm);
	_tprintf(_T("Max absolute error: %E\n\n"), dMaxDelta);
	_tprintf((dL1Norm < 0.0001) ? _T("TEST PASSED\n\n") : _T("TEST FAILED\n\n"));

	CUT_SAFE_CALL(cutDeleteTimer(hTimer));
	delete pSourceData;
	delete pDestDataGold;
	delete pDestDataCuda;

	return 0;
}

float RandFloat(float low, float high)
{
    float t = (float)rand() / (float)RAND_MAX;
    return (1.0f - t) * low + t * high;
}
