#include <stdio.h>
#include <cutil.h>
#include <tchar.h>

#define DISPATCH_POSITION()\
	const int nNumThreadsInBlock = blockDim.x * blockDim.y * blockDim.z;\
	const int nNumBlocksInGrid = gridDim.x * gridDim.y;\
	const int nNumThreads = nNumBlocksInGrid * nNumThreadsInBlock;\
	const int nBlockIDInGrid = blockIdx.x + blockIdx.y * gridDim.x;\
	const int nThreadIDInBlock = threadIdx.x + threadIdx.y * blockDim.x +\
		threadIdx.z * blockDim.x * blockDim.y;\
	const int nThreadID = nBlockIDInGrid * nNumThreadsInBlock + nThreadIDInBlock;

__global__ void SMACudaKernel(float* pOut, float* pIn, int nPeriod, long lSize)
{
	DISPATCH_POSITION();
	//_tprintf(_T("Thread ID: %d, Total num threads in grid: %d\n"), nThreadID, nNumThreads);

	for (int nBar = nThreadID; nBar < lSize; nBar += nNumThreads)
	{
		int nStartBar = (nBar < nPeriod ? 0 : nBar - nPeriod + 1);

		float fSum = 0.0f;
		for (int i = nStartBar; i <= nBar; i++)
			fSum += pIn[i];
		pOut[nBar] = fSum / float(nBar - nStartBar + 1);
	}
}

extern "C" void SMACuda(float* pOut, float* pIn, float fPeriod, long lSize)
{
	float* d_pIn = NULL;
	float* d_pOut = NULL;

	CUDA_SAFE_CALL(cudaMalloc((void**)&d_pIn, lSize * sizeof(float)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&d_pOut, lSize * sizeof(float)));
	CUDA_SAFE_CALL(cudaMemcpy(d_pIn, pIn, lSize * sizeof(float), cudaMemcpyHostToDevice));

	unsigned int hTimer;
	CUT_SAFE_CALL(cutCreateTimer(&hTimer));
	CUT_SAFE_CALL(cutResetTimer(hTimer));
	CUT_SAFE_CALL(cutStartTimer(hTimer));

	CUDA_SAFE_CALL(cudaThreadSynchronize());
	SMACudaKernel<<<512, 512>>>(d_pOut, d_pIn, fPeriod, lSize);
	//SMACudaKernel<<<int(lSize / 512) + 1, 512>>>(d_pOut, d_pIn, fPeriod, lSize);
	CUDA_SAFE_CALL(cudaThreadSynchronize());

	CUT_SAFE_CALL(cutStopTimer(hTimer));
	double dGPUPureTime = cutGetTimerValue(hTimer);
    _tprintf(_T("GPU pure time: %f msec\n"), dGPUPureTime);

	CUDA_SAFE_CALL(cudaMemcpy(pOut, d_pOut, lSize * sizeof(float), cudaMemcpyDeviceToHost));
	CUDA_SAFE_CALL(cudaFree(d_pIn));
	CUDA_SAFE_CALL(cudaFree(d_pOut));
	CUT_SAFE_CALL(cutDeleteTimer(hTimer));
}
